/*
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  Author: Massimiliano Culpo (m.culpo@cineca.it)
*/

#ifndef MATRIX_MULTIPLY_BENCHMARK_H_
#define MATRIX_MULTIPLY_BENCHMARK_H_

#include <stdlib.h>

#define DEFAULT_M 10
#define DEFAULT_N 10
#define DEFAULT_P 10
#define DEFAULT_MMTYPE 0
#define DEFAULT_CACHE_OBLIVIOUS 32

#define TILE_II 16
#define TILE_JJ 16
#define TILE_KK 64

typedef struct mmparameters_ {
  int mmtype;
  int cache_oblivious_base_size;
} mmparameters;

typedef struct dmatrix_ {
  size_t    nrows;
  size_t    ncols;
  double ** values;
} dmatrix;

/* Read command-line parameters */
void read_input(int argc, char** argv, mmparameters* opt, dmatrix* A, dmatrix* B, dmatrix* C);
/* Allocate a dmatrix if nrow and ncol are set */
void allocate_matrix(dmatrix* A);
/* Initialize a matrix for benchmark */
void initialize_matrix(dmatrix*A);
/* Benchmark test */
void benchmark_matrix_multiplication(dmatrix* C, dmatrix* A, dmatrix * B, const mmparameters * opt);
/* Free memory */
void free_matrix(dmatrix* A);
/* Print matrix for debug purpose */
void print_matrix(const dmatrix* A);

#endif
