/*
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  Author: Massimiliano Culpo (m.culpo@cineca.it)
*/

#include <stdio.h>

#include "MatrixMultiplyBenchmark.h"
//#include "MatrixMultiplyTiming.h"
#include<time.h>
#include<time.h>

int main(int argc, char** argv) {

  dmatrix      A,B,C;
  mmparameters opt;

  /* Read command-line parameters */
  read_input(argc,argv,&opt,&A,&B,&C);

  /* Allocate matrices of correct size */
  allocate_matrix(&A);
  allocate_matrix(&B);
  allocate_matrix(&C);
  /* Initialize matrices */
  initialize_matrix(&A);
  initialize_matrix(&B);
#ifdef DEBUG
  printf("Matrix A:\n");
  print_matrix(&A);
  printf("Matrix B:\n");
  print_matrix(&B);
#endif
  
  /* Actual Benchmark */
  printf("Starting benchmark\n");
  clock_t start = clock();
  benchmark_matrix_multiplication(&C,&A,&B,&opt);  
  clock_t stop=clock();
  double ela=((stop-start)*10)/CLOCKS_PER_SEC;
  printf("\nElapsed time: %f ms\n",ela);
#ifdef DEBUG
  printf("Matrix C:\n");
  print_matrix(&C);
#endif

  /* Free memory */
  free_matrix(&A);
  free_matrix(&B);
  free_matrix(&C);

  return 0;
}
