/*
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  Author: Massimiliano Culpo (m.culpo@cineca.it)
*/

#include <stdio.h>
#include <stdlib.h>
#include <linux/time.h>
#include <time.h>

#include "MatrixMultiplyTiming.h"

#define TIMING_CLOCK_ID CLOCK_PROCESS_CPUTIME_ID

clockmark_t timing_nsec(const clockmark_t* const start, const clockmark_t* const end) {
  return *end - *start;
}

double      timing_sec (const clockmark_t* const start, const clockmark_t* const end) {
  return (double) timing_nsec(start, end) * 1.0e-9;
}

clockmark_t timing_getmark(void) {
  struct timespec temp;
  uint64_t        nanos;

  /* Try the default clock first, if it fails,
     try again with CLOCK_MONOTONIC */
  int stat = clock_gettime(TIMING_CLOCK_ID, &temp);
  if (stat != 0) {
    stat = clock_gettime(CLOCK_MONOTONIC , &temp);
    if (stat != 0) {
      perror("timing_getmark()");
      exit(-1);
    }
  }
  nanos = temp.tv_nsec;
  nanos += ((uint64_t)temp.tv_sec) * 1000 * 1000 * 1000;
  return nanos;
}

