/*
 04/21/12 Driven Cavity non lineare con operatore di Boltzamann
         con polinomi di Hermite e secondo ordine nello spazio
         e griglia non uniforme

*/


#include <stdio.h>
#include <memory.h>
#include <time.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include<omp.h>



#define min(x,y)   x < y ? x : y

extern "C" double ddot_(int *n, double *dx, int *incx, double *dy,int *incy);
extern "C" int dgemv_(char *trans, int *m, int *n, double *alpha, double *a, int *lda, double *x, int *incx,double *beta, double *y, int *incy);



///////////////////////////////
// Simulation parameters     //
///////////////////////////////

#define PI  3.141592654f

#define N  5              // Do not change  
int NIT = 6;            // number of time steps
int NS=2;		// solution saving frequency
double dt=0.1;        // Do not change
double Vw=0.001;      // Do not change

double Lx  = 10.0;       // X domain dimension
double DxM = 1.0;     // Do not change
double Rx  = 1.0;     // Do not change
int NX  = 10;          // number of psatial nodes in the X direction

double Ly  = 10.0;       // Y domain dimension
double DyM = 1.0;     // Do not change
double Ry  = 1.0;     // rDo not change
int NY  = 10;         // number of psatial nodes in the Y direction

int N2=  N*N;
int N3 = N*N*N;
int N6 = N*N*N*N*N*N;


int AS, is=-1;
double Cw0, CwL;
// timing variables
double start0,start,end;


double *gPP, *gPM, *gMP, *gMM;


double *MyM, *MyP;
double *MxM, *MxP;
double  *D, *Vx, *Vy, *T, *Pxx, *Pyy, *Pxy, *Qx, *Qy;

double *C111, *C112, *C113, *C114, *C122, *C123, *C124, *C133, *C134, *C144;
double *W, *Z, *Wp, *Zp, *Wm,*Zm;
double *We, *Ze2, *WeZe1, *WeZe2,*maxwLP, *maxwLM;
double *Ze;
double *x, *dx,*y, *dy;
double *xApg, *yApg;
double *dXIPx, *dYIPx, *dZIPx, *dXIMx, *dYIMx, *dZIMx, *ddx, *ddxP, *ddxM;
double *dXIPy, *dYIPy, *dZIPy, *dXIMy, *dYIMy, *dZIMy, *ddy, *ddyP, *ddyM;

//MPI Variable
int myid,nprocs,ierr;
int master=0;
int submat_dim,i1,i2;
double mstart, mend;
MPI_Status status;



void allocate(){

	gPP=(double*)malloc(NY*NX*N3*sizeof(double));
	gPM=(double*)malloc(NY*NX*N3*sizeof(double));
	gMM=(double*)malloc(NY*NX*N3*sizeof(double));
	gMP=(double*)malloc(NY*NX*N3*sizeof(double));


	MxM=(double*)malloc(NY*sizeof(double));
	MxP=(double*)malloc(NY*sizeof(double));
	MyM=(double*)malloc(NX*sizeof(double));
	MyP=(double*)malloc(NX*sizeof(double));

	D=(double*)malloc(NY*NX*NS*sizeof(double));
	Vx=(double*)malloc(NY*NX*NS*sizeof(double));
	Vy=(double*)malloc(NY*NX*NS*sizeof(double));
	T=(double*)malloc(NY*NX*NS*sizeof(double));
	Pxx=(double*)malloc(NY*NX*NS*sizeof(double));
	Pyy=(double*)malloc(NY*NX*NS*sizeof(double));
	Pxy=(double*)malloc(NY*NX*NS*sizeof(double));
	Qx=(double*)malloc(NY*NX*NS*sizeof(double));
	Qy=(double*)malloc(NY*NX*NS*sizeof(double));
	Pxy=(double*)malloc(NY*NX*NS*sizeof(double));

	 submat_dim = ceil((double)N6/nprocs/N3)*N3;


	 if ( myid == 0 ) i1 = (nprocs-1) * submat_dim;
	 else i1 = (myid-1) *submat_dim;

	 if ( myid == 0 ) i2 = submat_dim + i1;
	 else i2 = submat_dim + i1 -1;

	 if ( i2 >= N6 ) i2 = N6-1;

	 printf("myid=%d i1=%d i2=%d submat_dim=%d\n", myid,i1,i2,submat_dim);

	//Allocate MPI CXX
	if(nprocs>1){
	C111=(double*)malloc(submat_dim*N3*sizeof(double));
	C112=(double*)malloc(submat_dim*N3*sizeof(double));
	C113=(double*)malloc(submat_dim*N3*sizeof(double));
	C114=(double*)malloc(submat_dim*N3*sizeof(double));
	C122=(double*)malloc(submat_dim*N3*sizeof(double));
	C123=(double*)malloc(submat_dim*N3*sizeof(double));
	C124=(double*)malloc(submat_dim*N3*sizeof(double));
	C133=(double*)malloc(submat_dim*N3*sizeof(double));
	C134=(double*)malloc(submat_dim*N3*sizeof(double));
	C144=(double*)malloc(submat_dim*N3*sizeof(double));
	}
	else{
		C111=(double*)malloc(N6*N3*sizeof(double));
		C112=(double*)malloc(N6*N3*sizeof(double));
		C113=(double*)malloc(N6*N3*sizeof(double));
		C114=(double*)malloc(N6*N3*sizeof(double));
		C122=(double*)malloc(N6*N3*sizeof(double));
		C123=(double*)malloc(N6*N3*sizeof(double));
		C124=(double*)malloc(N6*N3*sizeof(double));
		C133=(double*)malloc(N6*N3*sizeof(double));
		C134=(double*)malloc(N6*N3*sizeof(double));
		C144=(double*)malloc(N6*N3*sizeof(double));
	}

	W=(double*)malloc(N*sizeof(double));
	Z=(double*)malloc(N*sizeof(double));
	Wp=(double*)malloc(N*sizeof(double));
	Zp=(double*)malloc(N*sizeof(double));
	Wm=(double*)malloc(N*sizeof(double));
	Zm=(double*)malloc(N*sizeof(double));


	We=(double*)malloc(N3*sizeof(double));
	Ze2=(double*)malloc(N3*sizeof(double));
	WeZe1=(double*)malloc(N3*sizeof(double));
	WeZe2=(double*)malloc(N3*sizeof(double));
	maxwLP=(double*)malloc(N3*sizeof(double));
	maxwLM=(double*)malloc(N3*sizeof(double));

	Ze=(double*)malloc(3*N3*sizeof(double));

	x=(double*)malloc(NX*sizeof(double));
	dx=(double*)malloc(NX*sizeof(double));

	y=(double*)malloc(NY*sizeof(double));
	dy=(double*)malloc(NY*sizeof(double));

	xApg=(double*)malloc((NX-2)*sizeof(double));
	yApg=(double*)malloc((NY-2)*sizeof(double));

	dXIPx=(double*)malloc((NY*(NX-2))*sizeof(double));
	dYIPx=(double*)malloc((NY*(NX-2))*sizeof(double));
	dZIPx=(double*)malloc((NY*(NX-2))*sizeof(double));
	dXIMx=(double*)malloc((NY*(NX-2))*sizeof(double));
	dYIMx=(double*)malloc((NY*(NX-2))*sizeof(double));
	dZIMx=(double*)malloc((NY*(NX-2))*sizeof(double));
	ddx=(double*)malloc((NY*(NX-2))*sizeof(double));
	ddxP=(double*)malloc((NY*(NX-2))*sizeof(double));
	ddxM=(double*)malloc((NY*(NX-2))*sizeof(double));

	dXIPy=(double*)malloc((NX*(NY-2))*sizeof(double));
	dYIPy=(double*)malloc((NX*(NY-2))*sizeof(double));
	dZIPy=(double*)malloc((NX*(NY-2))*sizeof(double));
	dXIMy=(double*)malloc((NX*(NY-2))*sizeof(double));
	dYIMy=(double*)malloc((NX*(NY-2))*sizeof(double));
	dZIMy=(double*)malloc((NX*(NY-2))*sizeof(double));
	ddy=(double*)malloc((NX*(NY-2))*sizeof(double));
	ddyP=(double*)malloc((NX*(NY-2))*sizeof(double));
	ddyM=(double*)malloc((NX*(NY-2))*sizeof(double));

}

void deallocate(){

		free(gPP);
		free(gPM);
		free(gMM);
		free(gMP);

		free(D);
		free(MxM);
		free(MxP);
		free(MyM);
		free(MyP);
		free(Vx);
		free(Vy);
		free(T);
		free(Pxx);
		free(Pyy);
		free(Pxy);
		free(Qx);
		free(Qy);


		free(C111);
		free(C112);
		free(C113);
		free(C114);
		free(C122);
		free(C123);
		free(C124);
		free(C133);
		free(C134);
		free(C144);

		free(W);
		free(Z);
		free(Wp);
		free(Zp);
		free(Wm);
		free(Zm);

		free(We);
		free(Ze2);
		free(WeZe1);
		free(WeZe2);
		free(maxwLP);
		free(maxwLM);
		free(Ze);

		free(x);
		free(dx);

		free(y);
		free(dy);

		free(xApg);
		free(yApg);

		free(dXIPx);
		free(dYIPx);
		free(dZIPx);
		free(dXIMx);
		free(dYIMx);
		free(dZIMx);

		free(ddx);

		free(ddxP);

		free(ddxM);

		free(dXIPy);
		free(dYIPy);
		free(dZIPy);
		free(dXIMy);
		free(dYIMy);
		free(dZIMy);
		free(ddy);
		free(ddyP);
		free(ddyM);
}

void write_grid(){
	int AS=NIT/NS;

	//Only Master;
	if(myid==master){
	//Write Time discretization
	FILE *F_time = fopen("time_C.dat","w");
	for(int it=1;it<NIT;it++){
		if ((it+1) % AS == 0)
		fprintf(F_time,"%.17f \n",dt*it);
	}
	fclose(F_time);
	}


	//Write xData discretization

	for(int i=0;i<NX/2;i++) dx[i]=DxM*pow(Rx,NX/2-1-i);
	for(int i=NX/2;i<NX;i++) dx[i]=DxM*pow(Rx,i-NX/2);
	x[0]=dx[0]/2.0;
	for(int i=1;i<NX;i++) x[i]=x[i-1]+0.5*(dx[i]+dx[i-1]);

	//Only Master
	if(myid==master){
	FILE *F_xdata = fopen("x_C.dat","w");
	for(int i=0;i<NX;i++) fprintf(F_xdata,"%.17f %.17f \n",x[i],dx[i]);
	fclose(F_xdata);
	}

	//Write yData discretization

	for(int i=0;i<NY;i++) dy[i]=DyM*pow(Ry,i);
	y[0]=dy[0]/2.0;
	for(int i=1;i<NY;i++) y[i]=y[i-1]+0.5*(dy[i]+dy[i-1]);

	//Only Master
	if(myid==master){
	FILE *F_ydata = fopen("y_C.dat","w");
	for(int i=NY-1;i>=0;i--) fprintf(F_ydata,"%.17f %.17f \n",y[i],dy[i]);
	fclose(F_ydata);
	}
}

void write_result(){
	FILE *fD, *fVx,  *fVy, *fT, *fPxy, *fPxx, *fPyy, *fQx, *fQy;
	fD = fopen("D_C_MPI.dat","w");
	if(fD==NULL){
		printf("Error Opening file\n");
		exit(1);

	}
	fVx = fopen("Vx_C_MPI.dat","w");
	fVy = fopen("Vy_C_MPI.dat","w");
	fPxy = fopen("Pxy_C_MPI.dat","w");
	fPxx = fopen("Pxx_C_MPI.dat","w");
	fPyy= fopen("Pyy_C_MPI.dat","w");
	fQy = fopen("Qy_C_MPI.dat","w");
	fQx = fopen("Qx_C_MPI.dat","w");
	fT = fopen("T_C_MPI.dat","w");





for(int it=0;it<NS;it++){
	for(int j=NY-1;j>=0;j--){
		for(int i=0;i<NX;i++){

		fprintf(fD,"%.16f\n",D[j+i*NY+it*NY*NX]);
		fprintf(fVx,"%.17f\n",Vx[j+i*NY+it*NY*NX]);
		fprintf(fVy,"%.16f\n",Vy[j+i*NY+it*NY*NX]);
		fprintf(fPxy,"%.16f\n",Pxy[j+i*NY+it*NY*NX]);
		fprintf(fPxx,"%.16f\n",Pxx[j+i*NY+it*NY*NX]);
		fprintf(fPyy,"%.16f\n",Pyy[j+i*NY+it*NY*NX]);
		fprintf(fQy,"%.16f\n",Qy[j+i*NY+it*NY*NX]);
		fprintf(fQx,"%.16f\n",Qx[j+i*NY+it*NY*NX]);
		fprintf(fT,"%.16f\n",T[j+i*NY+it*NY*NX]);

 }
}
}



fclose(fD);
fclose(fVx);
fclose(fVy);
fclose(fPxy);
fclose(fPxx);
fclose(fPyy);
fclose(fQy);
fclose(fQx);
fclose(fT);

}
void load_data(char *name, double *C, int number){
	char a[256];
	printf("Load C%i \n", number);

	FILE *F_C = fopen(name,"r");


	if(nprocs>1){
	for(int n_blo=1;n_blo<=nprocs-1;n_blo++){
	for (int i=0;i<submat_dim;i++){
		for(int j=0;j<N3;j++){
			if (i*j==(submat_dim-1)*(N3-1)){
				fscanf(F_C,"%s\n", a);
				C[i+j*submat_dim]=atof(a);
			}
			else{
				fscanf(F_C,"%s ", a);
				C[i+j*submat_dim]=atof(a);
			}
		}

	}
	//Send to all process
     ierr=MPI_Send( C, submat_dim*N3, MPI_DOUBLE, n_blo,100+n_blo, MPI_COMM_WORLD);
	}
	for (int i=0;i<=(i2-i1);i++){
			for(int j=0;j<N3;j++){
				if (i*j==(submat_dim-1)*(N3-1)){
					fscanf(F_C,"%s\n", a);
					C[i+j*submat_dim]=atof(a);
				}
				else{
					fscanf(F_C,"%s ", a);
					C[i+j*submat_dim]=atof(a);
				}
			}
	}

	}
	else{
		for (int i=0;i<N6;i++){
				for(int j=0;j<N3;j++){
					if (i*j==(N6-1)*(N3-1)){
						fscanf(F_C,"%s\n", a);
						C[i+j*N6]=atof(a);
					}
					else{
						fscanf(F_C,"%s ", a);
						C[i+j*N6]=atof(a);

					}
				}
			}
	}
	fclose(F_C);

}

 void receive_data(double *C){

	ierr=MPI_Recv( C, submat_dim*N3, MPI_DOUBLE , 0,100+myid, MPI_COMM_WORLD, &status );
}

void load_pesi(char *name, double *C){
	FILE *F_C = fopen(name,"r");
	int count=0;
	char a[256];
	while (!feof(F_C)){
		fscanf(F_C,"%s\n",a);
		C[count]=atof(a);
		count+=1;
	}
	//Send data to all slaves
	if(nprocs>1){
		for(int i=1;i<=nprocs-1;i++) ierr=MPI_Send(C,N, MPI_DOUBLE, i,200+i, MPI_COMM_WORLD);
	}
	fclose(F_C);
}


void inizialize(){

	double *vec1;
	int dim=N3;
    int inc=1;
    int ind;

 for (int i=0;i<NY;i++){
	 for(int j=0;j<NX-2;j++){
		 dXIPx[i+j*NY]=1.0/(x[2+j]-x[j+1]);
		 dYIPx[i+j*NY]=1.0/(x[j+1]-x[j]);
		 if (j!=0){
			 xApg[j]=x[j-1];
		 }
		 xApg[0]=-x[0];
		 dZIPx[i+j*NY] = 1.0  / (x[j] - xApg[j]);
		


		 dXIMx[i+j*NY] = -1.0 / (x[j] - x[j+1]);
		 dYIMx[i+j*NY] = -1.0 / (x[j+1] - x[2+j]);
		 if (j!=NX-2) xApg[j]=x[3+j];
		 xApg[NX-3] = x[NX-1] + dx[NX-1];
		 dZIMx[i+j*NY] = -1.0 /(x[j+2]   - xApg[j]);

		 ddx[i+j*(NY)]  = dx[j+1];
		 ddxP[i+j*(NY)] = dx[j];
		 ddxM[i+j*(NY)] = dx[j+2];
	 }
 }


 for (int j=0;j<NY-2;j++){
 	 for(int i=0;i<NX;i++){

 		dXIPy[j+i*(NY-2)]=1.0/ (y[j+2] - y[j+1]);
 		dYIPy[j+i*(NY-2)] = 1.0 / (y[j+1] - y[j]);
 		if(j!=0) yApg[j] = y[j-1];
 		yApg[0] = - y[0];
 		dZIPy[j+i*(NY-2)] = 1.0 / (y[j]- yApg[j]);//NO




		dXIMy[j+i*(NY-2)] = -1.0 / (y[j] - y[j+1]);
		dYIMy[j+i*(NY-2)] = -1.0 / (y[j+1] - y[j+2]);
		if(j!=NY-2) yApg[j] = y[3+j];
		yApg[NY-3] = y[NY-1] + dy[NY-1];
		dZIMy[j+i*(NY-2)] = -1.0 / (y[j+2]  - yApg[j]);
		ddy[j+i*(NY-2)]  = dy[1+j];
		ddyP[j+i*(NY-2)] = dy[j];
		ddyM[j+i*(NY-2)]= dy[j+2];


 	 }
}


for(int i=0;i<NY;i++){
	for(int j=0;j<NX;j++){
		for(int k=0;k<N3;k++){
			gPP[i+j*NY+k*NX*NX]=1.0;
			gPM[i+j*NY+k*NX*NX]=1.0;
			gMP[i+j*NY+k*NX*NX]=1.0;
			gMM[i+j*NY+k*NX*NX]=1.0;
		}
		for(int k=0;k<NS;k++){
			D [i+j*NY+k*NX*NX]  = 0.0;
			Pxx [i+j*NY+k*NX*NX] = 0.0;
			Pyy[i+j*NY+k*NX*NX]  = 0.0;
			Pxy[i+j*NY+k*NX*NX]  = 0.0;
			Qx[i+j*NY+k*NX*NX]   = 0.0;
			Qy[i+j*NY+k*NX*NX]   = 0.0;
			Vx[i+j*NY+k*NX*NX]  = 0.0;
			Vy[i+j*NY+k*NX*NX]  = 0.0;
			T[i+j*NY+k*NX*NX]   = 0.0;

		}
	}
}
Cw0=0.0;
CwL=0.0;
for(int k=0;k<N;k++){
	for(int j=0;j<N;j++){
		for(int i=0;i<N;i++){

			  ind=i+j*N+k*N2;

			  We[ind] = W[k]*Wp[j]*Wp[i];
			  Ze[ind] = Zp[i];
			  Ze[ind+N3] = Zp[j];
		      Ze[ind+2*N3] = Z[k];

		      Ze2[ind]=Zp[i]*Zp[i]+Zp[j]*Zp[j] + Z[k]*Z[k];

		      maxwLP[ind] = exp( Ze[ind]*Vw - 0.50*pow(Vw,2));
		      maxwLM[ind] = exp(-Ze[ind]*Vw - 0.50*pow(Vw,2));
		      Cw0 +=  2.0*We[ind]*Ze[ind];
		      CwL +=  We[ind]*(maxwLP[ind]+maxwLM[ind])*Ze[ind+N3];
		}
	}
}

// inizializzazione flussi incidenti le quattro pareti
for(int i=0;i<N3;i++){
WeZe1[i] =We[i]*Ze[i];
WeZe2[i] =We[i]*Ze[i+N3];
}


vec1=(double*)malloc(N3*sizeof(double));
// in y = 0, L
for(int i=0;i<NX;i++){
	for(int k=0;k<N3;k++){
		vec1[k]=gPM[0+i*NY+k*NX*NY]+gMM[0+i*NY+k*NX*NY];
	}
	MyM[i]=ddot_(&dim,vec1,&inc,WeZe2, &inc);
	for(int k=0;k<N3;k++){
			vec1[k]=gPP[NY-1+i*NY+k*NX*NY]+gMP[NY-1+i*NY+k*NX*NY];
		}
	MyP[i]=ddot_(&dim,vec1,&inc,WeZe2, &inc);
}
free(vec1);



vec1=(double*)malloc(N3*sizeof(double));
// in x = 0, L
for(int j=0;j<NY;j++){
	for(int k=0;k<N3;k++){
		vec1[k]=gMP[j+k*NX*NY]+gMM[j+k*NX*NY];

	}
	MxM[j]=ddot_(&dim,vec1,&inc,WeZe1, &inc);
	for(int k=0;k<N3;k++){
			vec1[k]=gPP[j+(NX-1)*NY+k*NX*NY]+gMP[j+(NX-1)*NY+k*NX*NY];

		}
	MxP[j]=ddot_(&dim,vec1,&inc,WeZe1, &inc);
}
free(vec1);


}


 double signF(double Z)
{
	if (Z >= 0.0)	return 1.0;
	if (Z < 0.0)	return -1.0;
}


void streaming(int it){
	double *gp1x, *gx, *gm1x, *gm2x, *XIx, *YIx, *ZIx, *minmodPx, *minmodMx,*Cx;
	double *gp1y, *gy, *gm1y, *gm2y, *XIy, *YIy, *ZIy, *minmodPy, *minmodMy,*Cy;

	//Allocate
	gp1x=(double*)malloc((NY*(NX-2))*sizeof(double));
	gx=(double*)malloc((NY*(NX-2))*sizeof(double));
	gm1x=(double*)malloc((NY*(NX-2))*sizeof(double));
	gm2x=(double*)malloc((NY*(NX-2))*sizeof(double));
	XIx=(double*)malloc((NY*(NX-2))*sizeof(double));
	YIx=(double*)malloc((NY*(NX-2))*sizeof(double));
	ZIx=(double*)malloc((NY*(NX-2))*sizeof(double));
	minmodPx=(double*)malloc((NY*(NX-2))*sizeof(double));
	minmodMx=(double*)malloc((NY*(NX-2))*sizeof(double));
	Cx=(double*)malloc((NY*(NX-2))*sizeof(double));

	gp1y=(double*)malloc((NX*(NY-2))*sizeof(double));
	gy=(double*)malloc((NX*(NY-2))*sizeof(double));
	gm1y=(double*)malloc((NX*(NY-2))*sizeof(double));
	gm2y=(double*)malloc((NX*(NY-2))*sizeof(double));
	XIy=(double*)malloc((NX*(NY-2))*sizeof(double));
	YIy=(double*)malloc((NX*(NY-2))*sizeof(double));
	ZIy=(double*)malloc((NX*(NY-2))*sizeof(double));
	minmodPy=(double*)malloc((NX*(NY-2))*sizeof(double));
	minmodMy=(double*)malloc((NX*(NY-2))*sizeof(double));
	Cy=(double*)malloc((NX*(NY-2))*sizeof(double));


	//Convection phase

        // X direction
	for(int k=0;k<N3;k++){
		
		for(int j=0;j<NY;j++){
			for(int i=0;i<NX-2;i++){
		

		Cx[j+i*NY] = dt/ddx[j+i*NY] * Ze[k];
		gp1x[j+i*NY] = gPP[j+(i+2)*NY+k*NY*NX];
		gx[j+i*NY]=gPP[j+(i+1)*NY+k*NY*NX];
		gm1x[j+i*NY]=gPP[j+i*NY+k*NY*NX];

		if(i<NX-3) gm2x[j+(i+1)*NY] = gPP[j+i*NY+k*NY*NX];
		gm2x[j+0*NY] = MxM[j]/Cw0;
		XIx[j+i*NY] = (gp1x[j+i*NY] - gx[j+i*NY])   * dXIPx[j+i*NY]+dXIPx[j+i*NY];
		YIx[j+i*NY] = (gx[j+i*NY] - gm1x[j+i*NY])   * dYIPx[j+i*NY]+dYIPx[j+i*NY];
		ZIx[j+i*NY] = (gm1x[j+i*NY] - gm2x[j+i*NY]) * dZIPx[j+i*NY]+dZIPx[j+i*NY];

		minmodPx[j+i*NY] = (0.5*(signF(XIx[j+i*NY])+signF(YIx[j+i*NY])))*(min(fabs(XIx[j+i*NY]),fabs(YIx[j+i*NY])));
		minmodMx[j+i*NY] = (0.5*(signF(YIx[j+i*NY])+signF(ZIx[j+i*NY])))*(min(fabs(YIx[j+i*NY]),fabs(ZIx[j+i*NY])));
		}
		}

		//Update GPP
		for(int j=0;j<NY;j++){
			for(int i=0;i<NX-2;i++){
		gPP[j+(i+1)*NY+k*NY*NX] =gx[j+i*NY]-Cx[j+i*NY]*(gx[j+i*NY]+gm1x[j+i*NY]+0.5*((ddx[j+i*NY] -dt*Ze[k])*minmodPx[j+i*NY]-
					(ddxP[j+i*NY]+dt*Ze[k])*minmodMx[j+i*NY]));

		
		gPP[j+k*NY*NX]=(1.0-Ze[k]*dt/dx[0])*gm1x[j]+Ze[k]*dt/dx[0];
		gPP[j+(NX-1)*NY+k*NY*NX]=gp1x[j+(NX-3)*NY] - Ze[k]*dt/dx[NX-1]*gx[j+(NX-3)*NY];
		}
		}
		
		for(int j=0;j<NY;j++){
			for(int i=0;i<NX-2;i++){
				Cx[j+i*NY] = dt/ddx[j+i*NY] * Ze[k];
				gp1x[j+i*NY] = gMP[j+i*NY+k*NY*NX];
				gx[j+i*NY]   = gMP[j+(i+1)*NY+k*NY*NX];
				gm1x[j+i*NY] = gMP[j+(i+2)*NY+k*NY*NX];
				gm2x[j+i*NY] = gMP[j+(i+3)*NY+k*NY*NX];
				
				gm2x[j+(NX-3)*NY] =  MxP[j]/Cw0;

				XIx[j+i*NY] = (gp1x[j+i*NY] - gx[j+i*NY])   * dXIMx[j+i*NY]-dXIMx[j+i*NY];
				YIx[j+i*NY] = (gx[j+i*NY]   - gm1x[j+i*NY]) * dYIMx[j+i*NY]-dYIMx[j+i*NY];
				ZIx[j+i*NY] = (gm1x[j+i*NY] - gm2x[j+i*NY]) * dZIMx[j+i*NY]-dZIMx[j+i*NY];

	        minmodPx[j+i*NY] = (0.5*(signF(XIx[j+i*NY])+signF(YIx[j+i*NY])))*(min(fabs(XIx[j+i*NY]),fabs(YIx[j+i*NY])));
	        minmodMx[j+i*NY] = (0.5*(signF(YIx[j+i*NY])+signF(ZIx[j+i*NY])))*(min(fabs(YIx[j+i*NY]),fabs(ZIx[j+i*NY])));

		}
		}
		// Update gMP
		for(int j=0;j<NY;j++){
			for(int i=0;i<NX-2;i++){
	        gMP[j+(i+1)*NY+k*NY*NX] = gx[j+i*NY] - Cx[j+i*NY] *(gx[j+i*NY] + gm1x[j+i*NY]+ 0.5 * ((ddx[j+i*NY] -dt*Ze[k]) * minmodPx[j+i*NY]
	                                  -(ddxM[j+i*NY]-dt*Ze[k]) * minmodMx[j+i*NY]));

               
            
	        gMP[j+k*NX*NX]  = (1.0-Ze[k]*dt/dx[1])*gp1x[j]+ Ze[k]*dt/dx[0];
	        gMP[j+(NX-1)*NY+k*NY*NX] = gm1x[j+(NX-3)*NY]+ Ze[k]*dt/dx[NX-1]*gm2x[j+(NX-3)*NY];
		}
		}

		
		for(int j=0;j<NY;j++){
			for(int i=0;i<NX-2;i++){
		Cx[j+i*NY] = dt/ddx[j+i*NY] * Ze[k];
		gp1x[j+i*NY] = gMM[j+i*NY+k*NY*NX];
		gx[j+i*NY]   = gMM[j+(i+1)*NY+k*NY*NX];
		gm1x[j+i*NY] = gMM[j+(i+2)*NY+k*NY*NX];
		gm2x[j+i*NY] = gMM[j+(i+3)*NY+k*NY*NX];

		
		gm2x[j+(NX-3)*NY] =  MxP[j]/Cw0;

		XIx[j+i*NY] = (gp1x[j+i*NY] - gx[j+i*NY])   * dXIMx[j+i*NY]+dXIMx[j+i*NY];
		YIx[j+i*NY] = (gx[j+i*NY]   - gm1x[j+i*NY]) * dYIMx[j+i*NY]+dYIMx[j+i*NY];
		ZIx[j+i*NY] = (gm1x[j+i*NY] - gm2x[j+i*NY]) * dZIMx[j+i*NY]+dZIMx[j+i*NY];

		minmodPx[j+i*NY] = (0.5*(signF(XIx[j+i*NY])+signF(YIx[j+i*NY])))*(min(fabs(XIx[j+i*NY]),fabs(YIx[j+i*NY])));
		minmodMx[j+i*NY] = (0.5*(signF(YIx[j+i*NY])+signF(ZIx[j+i*NY])))*(min(fabs(YIx[j+i*NY]),fabs(ZIx[j+i*NY])));
		}
		}
		
                // Update gMM
		for(int j=0;j<NY;j++){
			for(int i=0;i<NX-2;i++){
		gMM[j+(i+1)*NY+k*NY*NX]= gx[j+i*NY] - Cx[j+i*NY] *(gx[j+i*NY] + gm1x[j+i*NY]+ 0.5 * ((ddx[j+i*NY] -dt*Ze[k]) * minmodPx[j+i*NY]
					 					  -(ddxM[j+i*NY]-dt*Ze[k]) * minmodMx[j+i*NY]));
		
		gMM[j+k*NY*NX] = (1.0-Ze[k]*dt/dx[1])*gp1x[j]+ Ze[k]*dt/dx[0];
		gMM[j+(NX-1)*NY+k*NY*NX] = gm1x[j+(NX-3)*NY]+ Ze[k]*dt/dx[NX-1]*gm2x[j+(NX-3)*NY];
		}
		}
		
		
		for(int j=0;j<NY;j++){
			for(int i=0;i<NX-2;i++){
		Cx[j+i*NY] = dt/ddx[j+i*NY] * Ze[k];
		gp1x[j+i*NY] = gPM[j+(i+2)*NY+k*NY*NX];
		gx[j+i*NY]   = gPM[j+(i+1)*NY+k*NY*NX];
		gm1x[j+i*NY] = gPM[j+i*NY+k*NY*NX];
		if(i<NX-3) gm2x[j+(i+1)*NY]=gPM[j+i*NY+k*NY*NX];
		
		gm2x[j] = MxM[j]/Cw0;
		XIx[j+i*NY] = (gp1x[j+i*NY] - gx[j+i*NY])   * dXIPx[j+i*NY]+dXIPx[j+i*NY];
		YIx[j+i*NY] = (gx[j+i*NY]   - gm1x[j+i*NY]) * dYIPx[j+i*NY]+dYIPx[j+i*NY];
		ZIx[j+i*NY] = (gm1x[j+i*NY] - gm2x[j+i*NY]) * dZIPx[j+i*NY]+dZIPx[j+i*NY];

		minmodPx[j+i*NY] = (0.5*(signF(XIx[j+i*NY])+signF(YIx[j+i*NY])))*(min(fabs(XIx[j+i*NY]),fabs(YIx[j+i*NY])));
		minmodMx[j+i*NY] = (0.5*(signF(YIx[j+i*NY])+signF(ZIx[j+i*NY])))*(min(fabs(YIx[j+i*NY]),fabs(ZIx[j+i*NY])));
		
		}
		}


		//Update gPM
		for(int j=0;j<NY;j++){
			for(int i=0;i<NX-2;i++){
		gPM[j+(i+1)*NY+k*NY*NX] =gx[j+i*NY]-Cx[j+i*NY]*(gx[j+i*NY]-gm1x[j+i*NY]+0.5*((ddx[j+i*NY] -dt*Ze[k])*minmodPx[j+i*NY]-(ddxP[j+i*NY]-dt*Ze[k])*minmodMx[j+i*NY]));

		
		gPM[j+k*NY*NX]  = (1.0-Ze[k]*dt/dx[0])*gm1x[j]+ Ze[k]*dt/dx[0];
		gPM[j+(NX-1)*NY+NY*NX*k] = gp1x[j+NY*(NX-3)] + Ze[k]*dt/dx[NX-1]*gx[j+NY*(NX-3)];
				}
				}



}
	
	        //Y Direction
		for(int k=0;k<N3;k++){
		 //LEGGO
		  for(int j=0;j<NY-2;j++){
			  for(int i=0;i<NX;i++){
		   // primo quadrante in y
		   Cy[j+i*(NY-2)] = dt/ddy[j+i*(NY-2)] * Ze[k+N3];
		   gp1y[j+i*(NY-2)] = gPP[j+2+i*NY+k*NX*NY];
     		   gy[j+i*(NY-2)]   = gPP[j+1+i*NY+k*NX*NY];
	           gm1y[j+i*(NY-2)] = gPP[j+i*NY+k*NX*NY];
		   if(j<NY-3) gm2y[j+1+i*(NY-2)] = gPP[j+i*NY+k*NX*NY];
	           gm2y[0+i*(NY-2)] = MyM[i]/Cw0;
                   XIy[j+i*(NY-2)] = (gp1y[j+i*(NY-2)] - gy[j+i*(NY-2)])   * dXIPy[j+i*(NY-2)]+dXIPy[j+i*(NY-2)];
                   YIy[j+i*(NY-2)] = (gy[j+i*(NY-2)]   - gm1y[j+i*(NY-2)]) * dYIPy[j+i*(NY-2)]+dYIPy[j+i*(NY-2)];
                   ZIy[j+i*(NY-2)] = (gm1y[j+i*(NY-2)] - gm2y[j+i*(NY-2)]) * dZIPy[j+i*(NY-2)]+dZIPy[j+i*(NY-2)];

		
		   minmodPy[j+i*(NY-2)]= (0.5*(signF(XIy[j+i*(NY-2)])+signF(YIy[j+i*(NY-2)])))*(min(fabs(XIy[j+i*(NY-2)]),fabs(YIy[j+i*(NY-2)])));
		   minmodMy[j+i*(NY-2)]= (0.5*(signF(YIy[j+i*(NY-2)])+signF(ZIy[j+i*(NY-2)])))*(min(fabs(YIy[j+i*(NY-2)]),fabs(ZIy[j+i*(NY-2)])));
							}
		  }

		  //Update gPP
		  for(int j=0;j<NY-2;j++){
		  	for(int i=0;i<NX;i++){
		   gPP[j+1+i*NY+k*NX*NY]= gy[j+i*(NY-2)] - Cy[j+i*(NY-2)] * ( gy[j+i*(NY-2)] + gm1y[j+i*(NY-2)]+0.5 * ((ddy[j+i*(NY-2)]-dt*Ze[k+N3])*minmodPy[j+i*(NY-2)]
		                                  -(ddyP[j+i*(NY-2)]-dt*Ze[k+N3])*minmodMy[j+i*(NY-2)]));
		   
		   gPP[i*NY+k*NX*NY] = (1.0-Ze[k+N3]*dt/dy[0]) *gm1y[i*(NY-2)]+ Ze[k+N3]*dt/dy[0];
		   gPP[NY-1+i*NY+k*NX*NY] = gp1y[NY-3+i*(NY-2)] + Ze[k+N3]*dt/dy[NY-1]*gy[NY-3+i*(NY-2)];
		  	
							}
		  }

		   
		  for(int j=0;j<NY-2;j++){
				for(int i=0;i<NX;i++){
					Cy[j+i*(NY-2)] = dt/ddy[j+i*(NY-2)] * Ze[k+N3];
				   gp1y[j+i*(NY-2)] = gMP[j+2+i*NY+k*NX*NY];
				   gy[j+i*(NY-2)]   = gMP[j+1+i*NY+k*NX*NY];
				   gm1y[j+i*(NY-2)] = gMP[j+i*NY+k*NX*NY];
				   if(j<NY-3) gm2y[j+1+i*(NY-2)] = gMP[j+i*NY+k*NX*NY];
				   
				   gm2y[i*(NY-2)] = MyM[i]/Cw0;

				   XIy[j+i*(NY-2)] = (gp1y[j+i*(NY-2)] - gy[j+i*(NY-2)])   * dXIPy[j+i*(NY-2)]+dXIPy[j+i*(NY-2)];
				   YIy[j+i*(NY-2)] = (gy[j+i*(NY-2)]   - gm1y[j+i*(NY-2)]) * dYIPy[j+i*(NY-2)]+dYIPy[j+i*(NY-2)];
				   ZIy[j+i*(NY-2)] = (gm1y[j+i*(NY-2)] - gm2y[j+i*(NY-2)]) * dZIPy[j+i*(NY-2)]+dZIPy[j+i*(NY-2)];

				   minmodPy[j+i*(NY-2)] = (0.5*(signF(XIy[j+i*(NY-2)])+signF(YIy[j+i*(NY-2)])))*(min(fabs(XIy[j+i*(NY-2)]),fabs(YIy[j+i*(NY-2)])));
				   minmodMy[j+i*(NY-2)] = (0.5*(signF(YIy[j+i*(NY-2)])+signF(ZIy[j+i*(NY-2)])))*(min(fabs(YIy[j+i*(NY-2)]),fabs(ZIy[j+i*(NY-2)])));

				}
		  }

		  //Update gMP
		  for(int j=0;j<NY-2;j++){
		  				for(int i=0;i<NX;i++){
		   gMP[j+1+i*NY+k*NX*NY]= gy[j+i*(NY-2)] - Cy[j+i*(NY-2)] * ( gy[j+i*(NY-2)] - gm1y[j+i*(NY-2)]+ 0.5 * ((ddy[j+i*(NY-2)] -dt*Ze[k+N3])*minmodPy[j+i*(NY-2)]
									   -(ddyP[j+i*(NY-2)]-dt*Ze[k+N3])*minmodMy[j+i*(NY-2)]));
		   gMP[i*NY+k*NX*NY]  = (1.0-Ze[k+N3]*dt/dy[0]) *gm1y[i*(NY-2)]+ Ze[k+N3]*dt/dy[0];
		   gMP[NY-1+i*NY+k*NX*NY] = gp1y[NY-3+i*(NY-2)] + Ze[k+N3]*dt/dy[NY-1]*gy[NY-3+i*(NY-2)];

		  	}
		  }

		   
		for(int j=0;j<NY-2;j++){
				for(int i=0;i<NX;i++){
			 
			 Cy[j+i*(NY-2)] = dt/ddy[j+i*(NY-2)] * Ze[k+N3];
			 gp1y[j+i*(NY-2)] = gMM[j+i*NY+k*NX*NY];
			 gy[j+i*(NY-2)]   = gMM[j+1+i*NY+k*NX*NY];
			 gm1y[j+i*(NY-2)] = gMM[j+2+i*NY+k*NX*NY];

			 if (j<NY-3)
				 gm2y[j+(NY-2)*i] = gMM[j+3+i*NY+k*NX*NY];
			 if (j==NY-3)
			 //gm2y for j=NY-3
				gm2y[j+(NY-2)*i] = MyP[i]/CwL * maxwLM[k];

			 if (j<NY-3) yApg[j] = y[j+3];
			 else yApg[NY-3] = y[NY-1] + dy[NY-1];

			 XIy[j+i*(NY-2)] = (gp1y[j+i*(NY-2)] - gy[j+i*(NY-2)])   * dXIMy[j+i*(NY-2)]+dXIMy[j+i*(NY-2)];
			 YIy[j+i*(NY-2)] = (gy[j+i*(NY-2)]   - gm1y[j+i*(NY-2)]) * dYIMy[j+i*(NY-2)]+dYIMy[j+i*(NY-2)];
			 ZIy[j+i*(NY-2)] = (gm1y[j+i*(NY-2)] - gm2y[j+i*(NY-2)]) * dZIMy[j+i*(NY-2)]+dZIMy[j+i*(NY-2)];

			 minmodPy[j+i*(NY-2)] = (0.5*(signF(XIy[j+i*(NY-2)])+signF(YIy[j+i*(NY-2)])))*(min(fabs(XIy[j+i*(NY-2)]),fabs(YIy[j+i*(NY-2)])));
			 minmodMy[j+i*(NY-2)] =(min(fabs(YIy[j+i*(NY-2)]),fabs(ZIy[j+i*(NY-2)])))*(0.5*(signF(YIy[j+i*(NY-2)])+signF(ZIy[j+i*(NY-2)])));

				}
		}

		//Update gMM
		for(int j=0;j<NY-2;j++){
				for(int i=0;i<NX;i++){
			    gMM[j+1+i*NY+k*NX*NY]=gy[j+i*(NY-2)]- Cy[j+i*(NY-2)]*(gy[j+i*(NY-2)]-gm1y[j+i*(NY-2)]+ 0.5*((ddy[j+i*(NY-2)]-dt*Ze[k+N3])*minmodPy[j+i*(NY-2)]
							             -(ddyM[j+i*(NY-2)] -dt*Ze[k+N3])*minmodMy[j+i*(NY-2)]));

			 
			 gMM[0+i*NY+k*NX*NY] = (1.0-Ze[k+N3]*dt/dy[0])*gp1y[i*(NY-2)]+ Ze[k+N3]*dt/dy[0];	 
			 gMM[NY-1+i*NY+k*NX*NY] =gm1y[NY-3+i*(NY-2)] + Ze[k+N3]*dt/dy[NY-1]*gm2y[NY-3+i*(NY-2)];
			 
				}
		}

	    

		for(int j=0;j<NY-2;j++){
				for(int i=0;i<NX;i++){
			   Cy[j+i*(NY-2)] = dt/ddy[j+i*(NY-2)] * Ze[k+N3];
			   gp1y[j+i*(NY-2)] = gPM[j+i*NY+k*NY*NX];
			   gy[j+i*(NY-2)]   = gPM[j+1+i*NY+k*NY*NX];
			   gm1y[j+i*(NY-2)] = gPM[j+2+i*NY+k*NY*NX];

			   if (j<NY-3) gm2y[j+i*(NY-2)]=gPM[j+3+i*NY+k*NY*NX];
			   
			   gm2y[NY-3+i*(NY-2)] = MyP[i]/CwL * maxwLP[k];
			   XIy[j+i*(NY-2)] = (gp1y[j+i*(NY-2)] - gy[j+i*(NY-2)])   * dXIMy[j+i*(NY-2)]+dXIMy[j+i*(NY-2)];
			   YIy[j+i*(NY-2)] = (gy[j+i*(NY-2)]   - gm1y[j+i*(NY-2)]) * dYIMy[j+i*(NY-2)]+dYIMy[j+i*(NY-2)];
			   ZIy[j+i*(NY-2)] = (gm1y[j+i*(NY-2)] - gm2y[j+i*(NY-2)]) * dZIMy[j+i*(NY-2)]+dZIMy[j+i*(NY-2)];

			   minmodPy[j+i*(NY-2)] = (0.5*(signF(XIy[j+i*(NY-2)])+signF(YIy[j+i*(NY-2)])))*(min(fabs(XIy[j+i*(NY-2)]),fabs(YIy[j+i*(NY-2)])));
			   minmodMy[j+i*(NY-2)] =(0.5*(signF(YIy[j+i*(NY-2)])+signF(ZIy[j+i*(NY-2)])))*(min(fabs(YIy[j+i*(NY-2)]),fabs(ZIy[j+i*(NY-2)])));
			   
			}
		}



		//Update gPM
		for(int j=0;j<NY-2;j++){
				for(int i=0;i<NX;i++){
		
			   gPM[j+1+i*NY+k*NY*NX]=gy[j+i*(NY-2)] - Cy[j+i*(NY-2)] * ( gy[j+i*(NY-2)] - gm1y[j+i*(NY-2)]+ 0.5 * ((ddy[j+i*(NY-2)]-dt*Ze[k+N3])*minmodPy[j+i*(NY-2)]
							   -(ddyM[j+i*(NY-2)]-dt*Ze[k+N3])*minmodMy[j+i*(NY-2)]));
			
			   gPM[i*NY+k*NY*NX]  = (1.0-Ze[k+N3]*dt/dy[0])*gp1y[i*(NY-2)]+ Ze[k+N3]*dt/dy[0];
			   gPM[NY-1+i*NY+k*NY*NX] = gm1y[NY-3+i*(NY-2)] + Ze[k+N3]*dt/dy[NY-1]*gm2y[NY-3+i*(NY-2)];

		}
		}

	}

	free(gp1x); free(gx); free(gm1x); free(gm2x); free(XIx); free(YIx); free(ZIx); free(minmodPx); free(minmodMx);free(Cx);
	free(gp1y); free(gy); free(gm1y); free(gm2y); free(XIy); free(YIy); free(ZIy); free(minmodPy); free(minmodMy);free(Cy);

}


void collision(int it){
	   double* C;
	   double *gg1_m1, *gg2_m1, *gg3_m1, *gg4_m1, *gg1_p1, *gg2_p1, *gg3_p1, *gg4_p1;
	   double *ggPP, *ggMP, *ggMM, *ggPM;

	   int dim=N3;
	   int dim_4=4*N3;
	   double alpha=1.0;
	   double beta=0.0;

	   int inc=1;
	   double *vec4, *vec;
	   int i,j,k,ini,fin,l,m;
	   //MPI Variable
	   int stk, stk1,sub_N3, sub_N3rec;

	   //OpenMP Variable
	   int threads=1,thread_id=0;
	   
	   // Each prcess must fill only a subset of C matrices whic are filled in the loop for(k=0;k<sub_N3;k++){  
           // at line 1035
	   // The C matrices have dimension (4N3,4N3)
           // Computing of the subset for each process
	   if (myid!=0){
	       sub_N3=ceil((float)N3/nprocs);
	       stk=(myid-1)*sub_N3;
	   }
	   else{
	       sub_N3=ceil((float)(i2-i1)/N3);
	       stk=N3-sub_N3;
	   }

	   if (nprocs==1){
	        sub_N3=N3;
	        stk=0;
	   }

	   
	   vec=(double*)malloc(N3*sizeof(double));

	   ggPP=(double*)malloc(NY*NX*sizeof(double));
	   ggMP=(double*)malloc(NY*NX*sizeof(double));
	   ggMM=(double*)malloc(NY*NX*sizeof(double));
	   ggPM=(double*)malloc(NY*NX*sizeof(double));


	   //Open Parallel Region
	  	   #pragma omp parallel default(shared) private(i,j,k,l,m,ini,fin,C,gg1_m1, gg1_p1, gg2_m1,gg2_p1, gg3_m1,gg3_p1, gg4_m1,gg4_p1, vec4)
	  	   {
		   thread_id=omp_get_thread_num();
	  	   if (thread_id==0){
	  	          threads = omp_get_num_threads();
	  	          if (threads >1) {
	  	           if(myid==0) printf("Collision executed in hybrid mode with %d MPI processes and %d OpenMP threads per MPI process\n",nprocs,threads);
	  	          }else{
	  	            if (myid==0) printf(" Collision executed with %d  MPI processes\n",nprocs);
	  	          }
	  	   }

		   gg1_m1=(double*)malloc(4*N3*sizeof(double));
		   gg1_p1=(double*)malloc(4*N3*sizeof(double));
		   gg2_m1=(double*)malloc(4*N3*sizeof(double));
		   gg2_p1=(double*)malloc(4*N3*sizeof(double));
		   gg3_m1=(double*)malloc(4*N3*sizeof(double));
		   gg3_p1=(double*)malloc(4*N3*sizeof(double));
		   gg4_m1=(double*)malloc(4*N3*sizeof(double));
		   gg4_p1=(double*)malloc(4*N3*sizeof(double));
		   C=(double*)malloc(4*N3*4*N3*sizeof(double));
		   vec4=(double*)malloc(4*N3*sizeof(double));
	   //Parallelization of external loop
	   #pragma omp for
	   for(i=0; i<NX;i++){
	      for(j=0; j<NY;j++){
			   for(k=0;k<N3;k++){
	         // 
	         gg1_m1[k]      = gPP[j+i*NY+k*NY*NX]-1.0;
	         gg1_m1[k+N3]   = gMP[j+i*NY+k*NY*NX]-1.0;
	         gg1_m1[k+2*N3] = gMM[j+i*NY+k*NY*NX]-1.0;
	         gg1_m1[k+3*N3] = gPM[j+i*NY+k*NY*NX]-1.0;
	         gg1_p1[k] = gg1_m1[k] + 2.0;
	         gg1_p1[k+N3] = gg1_m1[k+N3] + 2.0;
	         gg1_p1[k+2*N3] = gg1_m1[k+2*N3] + 2.0;
	         gg1_p1[k+3*N3] = gg1_m1[k+3*N3] + 2.0;

	         // 
	         gg2_m1[k]      = gMP[j+i*NY+k*NY*NX]-1.0;
	         gg2_m1[k+N3]   = gPP[j+i*NY+k*NY*NX]-1.0;
	         gg2_m1[k+2*N3] = gPM[j+i*NY+k*NY*NX]-1.0;
	         gg2_m1[k+3*N3] = gMM[j+i*NY+k*NY*NX]-1.0;
	         gg2_p1[k] = gg2_m1[k] + 2.0;
	         gg2_p1[k+N3] = gg2_m1[k+N3] + 2.0;
	         gg2_p1[k+2*N3] = gg2_m1[k+2*N3] + 2.0;
	         gg2_p1[k+3*N3] = gg2_m1[k+3*N3] + 2.0;
	         //
	         gg3_m1[k]      = gMM[j+i*NY+k*NY*NX]-1.0;
	         gg3_m1[k+N3]   = gPM[j+i*NY+k*NY*NX]-1.0;
	         gg3_m1[k+2*N3] = gPP[j+i*NY+k*NY*NX]-1.0;
	         gg3_m1[k+3*N3] = gMP[j+i*NY+k*NY*NX]-1.0;
	         gg3_p1[k] = gg3_m1[k] + 2.0;
	         gg3_p1[k+N3] = gg3_m1[k+N3] + 2.0;
	         gg3_p1[k+2*N3] = gg3_m1[k+2*N3] + 2.0;
		     gg3_p1[k+3*N3] = gg3_m1[k+3*N3] + 2.0;
	         // 
	         gg4_m1[k]      = gPM[j+i*NY+k*NY*NX]-1.0;
	         gg4_m1[k+N3]   = gMM[j+i*NY+k*NY*NX]-1.0;
	         gg4_m1[k+2*N3] = gMP[j+i*NY+k*NY*NX]-1.0;
	         gg4_m1[k+3*N3] = gPP[j+i*NY+k*NY*NX]-1.0;
	         gg4_p1[k] = gg4_m1[k] + 2.0;
		     gg4_p1[k+N3] = gg4_m1[k+N3] + 2.0;
		     gg4_p1[k+2*N3] = gg4_m1[k+2*N3] + 2.0;
		     gg4_p1[k+3*N3] = gg4_m1[k+3*N3] + 2.0;
			 }
		  
	 	  // Fill of C matrices
		  for(k=0;k<sub_N3;k++){
				  ini=k*N3;
				  fin=ini+N3-1;
				  for(l=0;l<N3;l++){
					  for(m=0;m<N3;m++){
				  // first row coefficient matrix
				C[l+4*m*N3]        = C111[l+ini+m*submat_dim]+C112[(l+ini)+m*submat_dim];
				C[l+4*(m+N3)*N3]   = C112[(l+ini)+m*submat_dim]-C111[l+ini+m*submat_dim];
				C[l+4*(m+2*N3)*N3] = C113[(l+ini)+m*submat_dim]+C112[(l+ini)+m*submat_dim];
				C[l+4*(m+3*N3)*N3] = C114[(l+ini)+m*submat_dim]-C113[(l+ini)+m*submat_dim];

				// second row coefficient matrix
				C[(l+N3)+4*m*N3]       = C112[m+ini+l*submat_dim]+C112[m+ini+l*submat_dim];
				C[(l+N3)+4*(m+N3)*N3]   = C122[(l+ini)+m*submat_dim]-C112[m+ini+l*submat_dim];
				C[(l+N3)+4*(m+2*N3)*N3] = C123[(l+ini)+m*submat_dim]+C123[m+ini+l*submat_dim];
				C[(l+N3)+4*(m+3*N3)*N3] = C124[(l+ini)+m*submat_dim]-C123[m+ini+l*submat_dim];

				// third row coefficient matrix
				C[(l+2*N3)+4*m*N3]       = C113[m+ini+l*submat_dim]+C113[m+ini+l*submat_dim];
				C[(l+2*N3)+ 4*(m+N3)*N3]   = C123[m+ini+l*submat_dim]-C113[m+ini+l*submat_dim];
				C[(l+2*N3)+ 4*(m+2*N3)*N3] = C133[(l+ini)+m*submat_dim]+C123[m+ini+l*submat_dim];
				C[(l+2*N3)+ 4*(m+3*N3)*N3] = C134[(l+ini)+m*submat_dim]-C133[m+ini+l*submat_dim];

				// forth row coefficient matrix
				C[(l+3*N3)+ 4*m*N3 ]       = C114[m+ini+l*submat_dim]+C114[m+ini+l*submat_dim];
				C[(l+3*N3)+ 4*(m+N3)*N3]   = C124[m+ini+l*submat_dim]-C114[m+ini+l*submat_dim];
				C[(l+3*N3)+ 4*(m+2*N3)*N3] = C134[m+ini+l*submat_dim]+C124[m+ini+l*submat_dim];
				C[(l+3*N3)+ 4*(m+3*N3)*N3] = C144[(l+ini)+m*submat_dim]-C134[m+ini+l*submat_dim];

				  }
	            }

				// Update gPP,gMP,gMM,gPM

				dgemv_("N", &dim_4, &dim_4, &alpha,C, &dim_4, gg1_m1, &inc, &beta, vec4,&inc);
				gPP[j+i*NY+(k+stk)*NY*NX]+= dt * ddot_(&dim_4, gg1_p1,  &inc, vec4,&inc);
				
				dgemv_("N", &dim_4, &dim_4, &alpha,C, &dim_4, gg2_m1, &inc, &beta, vec4,  &inc);
				gMP[j+i*NY+(k+stk)*NY*NX]+=  dt * ddot_(&dim_4, gg2_p1,  &inc, vec4,&inc);

				dgemv_("N", &dim_4, &dim_4, &alpha,C, &dim_4, gg3_m1, &inc, &beta, vec4,  &inc);
				gMM[j+i*NY+(k+stk)*NY*NX]+=  dt *ddot_(&dim_4, gg3_p1,  &inc, vec4, &inc);

				dgemv_("N", &dim_4, &dim_4, &alpha,C, &dim_4, gg4_m1,  &inc, &beta, vec4,  &inc);
				gPM[j+i*NY+(k+stk)*NY*NX]+=  dt *ddot_(&dim_4, gg4_p1,  &inc, vec4, &inc);
			  }

		   }
	   } // End of i,j loops

	   free(vec4);
	   free(gg1_m1);
	   free(gg2_m1);
	   free(gg3_m1);
	   free(gg4_m1);
	   free(gg1_p1);
	   free(gg2_p1);
	   free(gg3_p1);
	   free(gg4_p1);
	   free(C);
//Close parallel region
}
// At the end of the loops in i,j the subsets of the arrays gPP,gPM,gMP,gMM owned by each process must be collected by the master 
// and broadcasted to all processes before calling the streaming function 
//All processes send to master process
	   if (nprocs>1){
	      if (myid !=0){

	          ierr=MPI_Send( &gPP[stk*NX*NY], NX*NY*sub_N3, MPI_DOUBLE, 0,1000+myid, MPI_COMM_WORLD);
	          ierr=MPI_Send( &gPM[stk*NX*NY], NX*NY*sub_N3, MPI_DOUBLE, 0,1000+myid, MPI_COMM_WORLD);
	          ierr=MPI_Send( &gMP[stk*NX*NY], NX*NY*sub_N3, MPI_DOUBLE, 0,1000+myid, MPI_COMM_WORLD);
	          ierr=MPI_Send( &gMM[stk*NX*NY], NX*NY*sub_N3, MPI_DOUBLE, 0,1000+myid, MPI_COMM_WORLD);
	      }
	      else{
// Master receives
	    	 sub_N3rec=ceil((double)N6/nprocs/N3);
	         for(int i=1;i<=nprocs-1;i++){
	           stk1=(i-1)*sub_N3rec;
	           ierr=MPI_Recv( &gPP[stk1*NX*NY], NX*NY*sub_N3rec, MPI_DOUBLE, i, MPI_ANY_TAG ,MPI_COMM_WORLD, &status);
	           ierr=MPI_Recv( &gPM[stk1*NX*NY], NX*NY*sub_N3rec, MPI_DOUBLE, i, MPI_ANY_TAG ,MPI_COMM_WORLD, &status);
	           ierr=MPI_Recv( &gMP[stk1*NX*NY], NX*NY*sub_N3rec, MPI_DOUBLE, i, MPI_ANY_TAG ,MPI_COMM_WORLD, &status);
	           ierr=MPI_Recv( &gMM[stk1*NX*NY], NX*NY*sub_N3rec, MPI_DOUBLE, i, MPI_ANY_TAG ,MPI_COMM_WORLD, &status);
	         }
	      }
// Master broadcasts to all processes
	    ierr=MPI_Bcast (gPP, NX*NY*N3, MPI_DOUBLE , 0, MPI_COMM_WORLD );
	    ierr=MPI_Bcast (gPM, NX*NY*N3, MPI_DOUBLE , 0, MPI_COMM_WORLD );
	    ierr=MPI_Bcast (gMP, NX*NY*N3, MPI_DOUBLE , 0, MPI_COMM_WORLD );
	    ierr=MPI_Bcast (gMM, NX*NY*N3, MPI_DOUBLE , 0, MPI_COMM_WORLD );

	   }


	      // y = 0, L
	      for(int i=0;i<NX;i++){
	    	  for(int k=0;k<N3;k++) vec[k]=gPM[0+i*NY+k*NX*NY] +gMM[0+i*NY+k*NX*NY];
	          MyM[i] = ddot_(&dim, vec, &inc, WeZe2,&inc);

	         for(int k=0;k<N3;k++) vec[k]=gPP[NY-1+i*NY+k*NX*NY] +gMP[NY-1+i*NY+k*NX*NY];
	         MyP[i] = ddot_(&dim, vec, &inc, WeZe2,&inc);
		  }
	      //x = 0, L
		  for(int j=0;j<NY;j++){
		  	 for(int k=0;k<N3;k++) vec[k]=gMP[j+0*NY+k*NX*NY] +gMM[j+0*NY+k*NX*NY];
		  	 MxM[j] = ddot_(&dim, vec,  &inc, WeZe1, &inc);

		  	 for(int k=0;k<N3;k++) vec[k]=gPP[j+(NX-1)*NY+k*NX*NY] +gPM[j+(NX-1)*NY+k*NX*NY];
	         MxP[j] = ddot_(&dim, vec,  &inc, WeZe1, &inc);

	      }

		 

		    //Only Master process
		    if(myid==0){
		    	int AS=NIT/NS;

		    if((it+1)% AS ==0){
		        is = is+1;

		        for(int k=0; k<N3;k++){
		           for(int j=0;j<NY;j++){
		        	   for(int i=0;i<NX;i++){

		           ggPP[j+i*NY] = gPP[j+i*NY+k*NX*NY]*We[k];
		           ggPM[j+i*NY] = gPM[j+i*NY+k*NX*NY]*We[k];
		           ggMP[j+i*NY] = gMP[j+i*NY+k*NX*NY]*We[k];
		           ggMM[j+i*NY] = gMM[j+i*NY+k*NX*NY]*We[k];


		           D[j+i*NY+is*NX*NY]   =  D[j+i*NY+is*NX*NY]+ggPP[j+i*NY]-ggPM[j+i*NY]+ggMP[j+i*NY]+ggMM[j+i*NY];
		           Vx[j+i*NY+is*NX*NY]  = Vx[j+i*NY+is*NX*NY]+(ggPP[j+i*NY]+ggPM[j+i*NY]-ggMP[j+i*NY]-ggMM[j+i*NY]) * Ze[k];
		           Vy[j+i*NY+is*NX*NY]  = Vy[j+i*NY+is*NX*NY]+ (ggPP[j+i*NY]+ggPM[j+i*NY]+ggMP[j+i*NY]-ggMM[j+i*NY]) * Ze[k+N3];
		           Pxx[j+i*NY+is*NX*NY] = Pxx[j+i*NY+is*NX*NY]+ (ggPP[j+i*NY]-ggPM[j+i*NY]+ggMP[j+i*NY]+ggMM[j+i*NY]) * Ze[k]*Ze[k];
		           Pyy[j+i*NY+is*NX*NY] = Pyy[j+i*NY+is*NX*NY]+ (ggPP[j+i*NY]+ggPM[j+i*NY]+ggMP[j+i*NY]+ggMM[j+i*NY]) * Ze[k+N3];
		           Pxy[j+i*NY+is*NX*NY] = Pxy[j+i*NY+is*NX*NY]+(ggPP[j+i*NY]-ggPM[j+i*NY]-ggMP[j+i*NY]+ggMM[j+i*NY]) * Ze[k];
		           T[j+i*NY+is*NX*NY]   = T[j+i*NY+is*NX*NY] + (ggPP[j+i*NY]+ggPM[j+i*NY]+ggMP[j+i*NY]+ggMM[j+i*NY]) * Ze2[k];
		           Qx[j+i*NY+is*NX*NY]  =  Qx[j+i*NY+is*NX*NY] +(ggPP[j+i*NY]+ggPM[j+i*NY]-ggMP[j+i*NY]-ggMM[j+i*NY]) * Ze2[k];
		           Qy[j+i*NY+is*NX*NY]  = Qy[j+i*NY+is*NX*NY]+ (ggPP[j+i*NY]-ggPM[j+i*NY]+ggMP[j+i*NY]-ggMM[j+i*NY]) * Ze2[k];

		           }
		           }

		        }

		        for(int j=0;j<NY;j++){
		        	for(int i=0;i<NX;i++){

					Vx[j+i*NY+is*NX*NY]  = Vx[j+i*NY+is*NX*NY]/D[j+i*NY+is*NX*NY];
					Vy[j+i*NY+is*NX*NY]  = Vy[j+i*NY+is*NX*NY]/D[j+i*NY+is*NX*NY];
					Pxx[j+i*NY+is*NX*NY] = Pxx[j+i*NY+is*NX*NY] - D[j+i*NY+is*NX*NY]*pow(Vx[j+i*NY+is*NX*NY],2);
					Pyy[j+i*NY+is*NX*NY]=  Pyy[j+i*NY+is*NX*NY]- D[j+i*NY+is*NX*NY]*pow(Vy[j+i*NY+is*NX*NY],2);
					Pxy[j+i*NY+is*NX*NY] =  Pxy[j+i*NY+is*NX*NY]- D[j+i*NY+is*NX*NY]*Vx[j+i*NY+is*NX*NY]*Vy[j+i*NY+is*NX*NY];
					Qx[j+i*NY+is*NX*NY]  = 0.50*Qx[j+i*NY+is*NX*NY]-Vx[j+i*NY+is*NX*NY]*Pxx[j+i*NY+is*NX*NY]-Vy[j+i*NY+is*NX*NY]*Pxy[j+i*NY+is*NX*NY]-0.5*Vx[j+i*NY+is*NX*NY]*T[j+i*NY+is*NX*NY];
					Qy[j+i*NY+is*NX*NY]  = 0.50*Qy[j+i*NY+is*NX*NY]-Vx[j+i*NY+is*NX*NY]*Pxy[j+i*NY+is*NX*NY]-Vy[j+i*NY+is*NX*NY]*Pyy[j+i*NY+is*NX*NY]-0.5*Vy[j+i*NY+is*NX*NY]*T[j+i*NY+is*NX*NY];
					T[j+i*NY+is*NX*NY] = 1.0/3.0/D[j+i*NY+is*NX*NY]*(T[j+i*NY+is*NX*NY] - D[j+i*NY+is*NX*NY]*(pow(Vx[j+i*NY+is*NX*NY],2)+pow(Vy[j+i*NY+is*NX*NY],2)));
		        	}
		        }
		    }
		    }


free(vec);

free(ggPP);
free(ggMP);
free(ggMM);
free(ggPM);


}

void write_C(char name[256], double *Cc){
	FILE *fCxxx;
	fCxxx=fopen(name ,"w");
    for(int i=0;i<N6;i++){
    	for(int j=0;j<N3;j++){
    		fprintf(fCxxx,"%.16f\n", Cc[i+j*N6]);
    	}
    	}
    fclose(fCxxx);

}

int main( int argc, char **argv)
{

	FILE *Fgmm, *Fgpp, *Fgmp, *Fgpm, *FgmmC, *FgppC, *FgmpC, *FgpmC;
	
	double elaS=0.0, elaC=0.0, elaTot=0.0;
	double startS, stopS, startC, stopC;

	//Start of parallel region

	ierr=MPI_Init(&argc, &argv);
	ierr=MPI_Comm_rank( MPI_COMM_WORLD, &myid );
	ierr=MPI_Comm_size( MPI_COMM_WORLD, &nprocs);

	printf("Allocate \n");
	allocate();
	printf("End allocating \n");

	printf("Write grid\n");
	write_grid();
	printf("End writing grid\n");

	//Only Master Read Data
	if (myid==0){
	load_data("/gpfs/scratch/userinternal/ainverni/DatiN5/C111_mod.dat", C111,111);
	load_data("/gpfs/scratch/userinternal/ainverni//DatiN5/C112_mod.dat", C112,112);
	load_data("/gpfs/scratch/userinternal/ainverni//DatiN5/C113_mod.dat", C113,113);
	load_data("/gpfs/scratch/userinternal/ainverni//DatiN5/C114_mod.dat", C114,114);
	load_data("/gpfs/scratch/userinternal/ainverni//DatiN5/C122_mod.dat", C122,122);
	load_data("/gpfs/scratch/userinternal/ainverni//DatiN5/C123_mod.dat", C123,123);
	load_data("/gpfs/scratch/userinternal/ainverni//DatiN5/C124_mod.dat", C124,124);
	load_data("/gpfs/scratch/userinternal/ainverni//DatiN5/C133_mod.dat", C133,133);
	load_data("/gpfs/scratch/userinternal/ainverni//DatiN5/C134_mod.dat", C134,134);
	load_data("/gpfs/scratch/userinternal/ainverni//DatiN5/C144_mod.dat", C144,144);
}
	//All processes Receive
	else{
	receive_data(C111);
	receive_data(C112);
	receive_data(C113);
	receive_data(C114);
	receive_data(C122);
	receive_data(C123);
	receive_data(C124);
	receive_data(C133);
	receive_data(C134);
	receive_data(C144);

	}


	


	if (myid==0){
	printf("Load pesi\n");
	load_pesi("/gpfs/scratch/userinternal/ainverni//Nodi_Pesi_Continui/Pesi_5", W);
	load_pesi("/gpfs/scratch/userinternal/ainverni//Nodi_Pesi_Continui/Zeri_5",Z);
	load_pesi("/gpfs/scratch/userinternal/ainverni//Nodi_Pesi_Discontinui/Pesi_5", Wp);
	load_pesi("/gpfs/scratch/userinternal/ainverni/Nodi_Pesi_Discontinui/Zeri_5", Zp);

	for (int i=0;i<N;i++){
		Wm[i]=-Wp[i];
		Zm[i]=-Zp[i];
	}
	for(int i=1;i<=nprocs-1;i++){
	ierr=MPI_Send(Wm,N, MPI_DOUBLE, i,200+i, MPI_COMM_WORLD);
	ierr=MPI_Send(Zm,N, MPI_DOUBLE, i,200+i, MPI_COMM_WORLD);
	}
	}
	else{
		ierr=MPI_Recv(W, N, MPI_DOUBLE ,  0,200+myid, MPI_COMM_WORLD, &status );
		ierr=MPI_Recv(Z, N, MPI_DOUBLE,  0,200+myid, MPI_COMM_WORLD, &status );
		ierr=MPI_Recv(Wp, N, MPI_DOUBLE ,  0,200+myid, MPI_COMM_WORLD, &status );
		ierr=MPI_Recv(Zp, N, MPI_DOUBLE ,  0,200+myid, MPI_COMM_WORLD, &status );
		ierr=MPI_Recv(Wm, N, MPI_DOUBLE ,  0,200+myid, MPI_COMM_WORLD, &status );
		ierr=MPI_Recv(Zm, N, MPI_DOUBLE,  0,200+myid, MPI_COMM_WORLD, &status );
	}

	printf("Inizialize\n");
	inizialize();

	for(int it=0;it<NIT-1;it++){
		if(myid==0)  printf("Iteration number %d\n",it);
		if(myid==0) {
                         startS=MPI_Wtime();
		         streaming(it);
                         stopS=MPI_Wtime();
                         elaS=(double)(stopS-startS);
   		         printf("Elapsed time streaming %f\n", elaS);
	                    } 
                         ierr=MPI_Bcast (gPP, NX*NY*N3, MPI_DOUBLE , 0, MPI_COMM_WORLD );
			 ierr=MPI_Bcast (gPM, NX*NY*N3, MPI_DOUBLE , 0, MPI_COMM_WORLD );
			 ierr=MPI_Bcast (gMP, NX*NY*N3, MPI_DOUBLE , 0, MPI_COMM_WORLD );
			 ierr=MPI_Bcast (gMM, NX*NY*N3, MPI_DOUBLE , 0, MPI_COMM_WORLD );


		if(myid==0) startC=MPI_Wtime();
		collision(it);
		if(myid==0) stopC=MPI_Wtime();
                elaC=(double)(stopC-startC); 
		if(myid==0) printf("Elapsed time collision %f\n", elaC);
		
		elaTot+=elaC+elaS;
	}
	if(myid==0)  printf("Total Elapsed Time = %f\n", elaTot);



	

    //Only Master
    if(myid==0){
    	 printf("Write result \n");
    	 write_result();
    }
	printf("Deallocate \n");
	deallocate();

	//MPI Finalize
	ierr=MPI_Finalize();
	return 0;
}
